// passchk.js is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation; either version 3 of the License, or (at your
// option) any later version.
//
// passchk.js is distributed in the hope that it will be useful but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// The passchk.js archive has a copy of the GNU General Public License,
// but if you did not get it, see <http://www.gnu.org/licenses/>
//
// passchk.js is available from http://rumkin.com/tools/password/passchk.php
//
// Javascript functions for the password checker form
//
// Revised by Kristofer Sweger, March 19, 2012:
//	Changed Common_Words to a Javascript Object to use faster associative lookups.
//	Changed timeouts from 200 to 300 milliseconds, which seems fast enough.
//	Report length, entropy and charset size last now on same line.
//	Changed output from using deprecated "<font color=..." to "<font style=...".
//	Search for common words throughout the password and issue warnings if found,
//		not just for a match to the entire password.
//	Check for passwords using only a small subset of character set and issue a warning.
//	Check for repeated (forward or reversed) character sequences and issue warnings.
//	Check for repeated single-character sequences and issue warnings.
//	Adjust (semi-arbitrarily?) entropy based on portion of password involved for common
//		words, character set subsets, and repeated characters/sequences.
//	Check for and issue caution messages when part of the password looks like a telephone 
//		number, a Social Security number, or a date. No adjustment is made to reported
//		entropy currently.
//
// Todo: Add street/address words/abbreviations to Common_Words or to a separate common.js table for cautions.
//

var Common_Words = new Object;
var Frequency_Table = new Array();
Frequency_Table[0] = 0;
var lastword = '';

// The compression algorithm is very basic - the first letter is upper case,
// and it means to copy X letters from the previous word.  A = 0, B = 1, etc.
// So, if I had "apple apricot banana", it would compress to
// "AappleCricotAbanana". 
function Parse_Common_Word()
{
   var i, c, word;
   i = 1;
   c = Common_List.substr(i, 1);
   while (c == c.toLowerCase() && i < Common_List.length)
   {
      i ++;
      c = Common_List.substr(i, 1);
   }

   word = Common_List.substr(0, i);
   Common_List = Common_List.substr(i, Common_List.length);

   if (word.substr(0, 1) == 'A')
   {
      word = word.substr(1, word.length);
   }
   else
   {
      i = word.charCodeAt(0) - 'A'.charCodeAt(0);
      word = lastword.substr(0, i) + word.substr(1, word.length);
   }
    Common_Words[word] = 1;
    lastword = word;
}

function Parse_Common()
{
   for (var i = 0; i < 100 && Common_List.length > 0; i ++)
   {
      Parse_Common_Word();
   }
   if (Common_List.length)
   {
      window.setTimeout('Parse_Common()', 20);
   }
   else
   {
      document.Common_Parsed = 1;
   }
}

// The frequency thing is a bit more interesting, but still not too complex.
// Each three letters are base-95 encoded number representing the chance that
// this combination comes next.  Subtract the value of ' ' from each of the
// three, then ((((first_value * 95) + second_value) * 95) + third_value) will
// give you the odds that this pair is grouped together.  The first is "  "
// (non-alpha chars), then " a", " b", etc. " y", " z", "a ", "aa", "ab", and
// so on.  If you decrypt the table successfully, you should see a really large
// number for "qu".
function Parse_Frequency_Token()
{
   var c;

   c = Frequency_List.charCodeAt(0) - ' '.charCodeAt(0);
   c /= 95;
   c += Frequency_List.charCodeAt(1) - ' '.charCodeAt(0);
   c /= 95;
   c += Frequency_List.charCodeAt(2) - ' '.charCodeAt(0);
   c /= 95;

   Frequency_List = Frequency_List.substr(3, Frequency_List.length);

   Frequency_Table[Frequency_Table.length] = c;
}


function Parse_Frequency()
{
   for (var i = 0; i < 100 && Frequency_List.length > 0; i ++)
   {
      Parse_Frequency_Token();
   }
   if (Frequency_List.length)
   {
      window.setTimeout('Parse_Frequency()', 20);
   }
   else
   {
      document.Frequency_Parsed = 1;
   }
}


function Get_Index(c)
{
   c = c.charAt(0).toLowerCase();
   if (c < 'a' || c > 'z')
   {
      return 0;
   }
   return c.charCodeAt(0) - 'a'.charCodeAt(0) + 1;
}


function Get_Charset_Size(pass)
{
   var a = 0, u = 0, n = 0, ns = 0, r = 0, sp = 0, s = 0, chars = 0;
   
   for (var i = 0 ; i < pass.length; i ++)
   {
      var c = pass.charAt(i);
      
      if (a == 0 && 'abcdefghijklmnopqrstuvwxyz'.indexOf(c) >= 0)
      {
         chars += 26;
	 a = 1;
      }
      if (u == 0 && 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'.indexOf(c) >= 0)
      {
         chars += 26;
	 u = 1;
      }
      if (n == 0 && '0123456789'.indexOf(c) >= 0)
      {
         chars += 10;
	 n = 1;
      }
      if (ns == 0 && '!@#$%^&*()'.indexOf(c) >= 0)
      {
         chars += 10;
	 ns = 1;
      }
      if (r == 0 && "`~-_=+[{]}\\|;:'\",<.>/?".indexOf(c) >= 0)
      {
         chars += 20;
	 r = 1;
      }
      if (sp == 0 && c == ' ')
      {
         chars += 1;
	 sp = 1;
      }
      if (s == 0 && (c < ' ' || c > '~'))
      {
         chars += 32 + 128;
	 s = 1;
      }
   }
   
   return chars;
}


function Set_Text(s)
{
   var e;

   if (! document.getElementById)
   {
      return;
   }

   e = document.getElementById('passchk_result');
   if (! e)
   {
      return;
   }

   if (e.innerHTML == s)
   {
      return;
   }

   e.innerHTML = s;
}


var OldPass = -1;
function ShowStats()
{
   var pass = document.passchk_form.passchk_pass.value;
   var plower = pass.toLowerCase();
   var r = "";

   if (pass == OldPass)
   {
      window.setTimeout('ShowStats();', 300);
      return;
   }
   OldPass = pass;

   if (pass.length == 0)
   {
      Set_Text("Enter a password to see its strength.");
      window.setTimeout('ShowStats();', 300);
      return;
   }

   if (pass.length <= 4)
   {
      r += "<b>WARNING: <font style='color:#c00;'>Very short password!</font></b><br/>\n";
   }
   else if (pass.length < 8)
   {
      r += "<b>WARNING: <font style='color:#c00;'>Short password!</font></b><br/>\n";
   }

   // First, see if it is a common password.
   pwords = plower.replace(/[^a-z]/g,' ');
   pwords = pwords.replace(/\ +/g,' ');
   parray = pwords.split(' ');
   commoncount = 0;
    if (Common_Words[plower] == 1)
      {
	 r += '<b>WARNING: <font style="color:#c00;">"'+plower+'" is a common password!</font></b><br/>\n';
	 commoncount += plower.length;
      }

    // See if the password contains any common words within it
    if (commoncount == 0)
    {
	repeats = '  ';
	for (var j = 0; j < parray.length; j++)
	{
	    ptemp = parray[j];
	    if (ptemp.length > 1)
	    {
		if (Common_Words[ptemp] == 1)
		{
		    if (repeats.indexOf(' '+ptemp+' ') < 1)
		    {
			r += '<b>WARNING: <font style="color:#c00;">Common word "'+ptemp+
			    '" is in the password!</font></b><br/>\n';
			commoncount += ptemp.length;
			repeats += ptemp+' ';
		    }
		}
	    }
	}
    }

    // Check unique characters in the password
    ptemp = "";
    for (var j = 0; j < pass.length; j++)
    {
	if (ptemp.indexOf(pass[j],0) < 0) ptemp += pass[j];
    }

    uniques = ptemp.length;

    if (uniques < 5)
    {
	if (uniques > 1)
	{
	    r += "<b>WARNING: <font style='color:#c00;'>Only "+uniques+" different characters used!</font></b><br/>\n";
	} else
	{
	    r += "<b>WARNING: <font style='color:#c00;'>Only one unique character used in password!</font></b><br/>\n";
	}
    }

    // Check for repeated sequences
    maxseqlen = Math.floor(pass.length / 2);
    repeats = ' ';
    repeatcount = 0;
    for (m = maxseqlen; m > 2; m--)
    {
	for (i = 0; i < pass.length - m; i++)
	{
	    seqstr = pass.substr(i,m);
	    if (pass.indexOf(seqstr,i+m) > 0)
	    {
		rtemp = Array(m+1).join(pass[i]);
		if (seqstr != rtemp && seqstr.indexOf(' ') < 0 && repeats.indexOf(seqstr) < 0)
		{
		    r += '<b>WARNING: <font style="color:#c00;">Sequence "'+seqstr+
			'" is repeated!</font></b><br/>\n';
		    repeatcount += m;
		    j = i+m;
		    k = pass.indexOf(seqstr,j);
		    while (k > 0)
		    {
			if (k == j) repeats += seqstr;
			repeatcount+=m;
			if (k > j) { j = k+m;} else {j += m;}
			k = pass.indexOf(seqstr,j);
		    }
		    repeats += seqstr + ' ';
		}
	    }
	}
    }

    // Check for reversed sequences
    maxseqlen = Math.floor(pass.length / 2);
    for (m = maxseqlen; m > 2; m--)
    {
	for (i = 0; i < pass.length - m; i++)
	{
	    seqstr = pass.substr(i,m);
	    splitstr = seqstr.split("");
	    reverstr = splitstr.reverse();
	    seqstr = reverstr.join("");
	    if (pass.indexOf(seqstr,i+m) > 0)
	    {
		rtemp = Array(m+1).join(pass[i]);
		if (seqstr != rtemp && seqstr.indexOf(' ') < 0 && repeats.indexOf(seqstr) < 1)
		{
		    r += '<b>WARNING: <font style="color:#c00;">Sequence "'+pass.substr(i,m)+
			'" is reversed in password!</font></b><br/>\n';
		    repeatcount += m;
		    j = i+m;
		    k = pass.indexOf(seqstr,j);
		    while (k > 0)
		    {
			if (k == j) repeats += seqstr;
			repeatcount+=m;
			if (k > j) { j = k+m;} else {j += m;}
			k = pass.indexOf(seqstr,j);
		    }
		    repeats += seqstr + ' ';
		}
	    }
	}
    }

    // Check for repeated single characters
    runchar = '';
    runcount = 0;
    runcur = 0;
    maxrun = 0;
    for (i = 0; i < pass.length; i++)
    {
	if (pass[i] != runchar)
	{
	    if (runcur > 0)
	    {
		if (runcur > 2)
		{
		    r += '<b>WARNING: <font style="color:#c00;">"'+runchar+
			'" is repeated '+runcur+' consecutive times!</font></b><br/>\n';
		    runcount += runcur;
		    if (runcur > maxrun) maxrun = runcur;
		}
	    }
	    runcur = 0;
	    runchar = pass[i];
	}
	runcur++;
    }
    if (runcur > 2)
    {
        r += '<b>WARNING: <font style="color:#c00;">"'+runchar+
	    '" is repeated '+runcur+' consecutive times!</font></b><br/>\n';
        runcount += runcur;
        if (runcur > maxrun) maxrun = runcur;
    }

    // Check for special numeric forms (date, SSN, phone number)
    var bioinfo = 0;

    // Telephone number
    var nondigit = /[^0-9]/;
    var pattern = /((\(\d{3}\)\ *)|(\d{3}(\-|\ +)))?\d{3}\-\d{4}/;
    if (pattern.test(pass))
    {
	phone = pattern.exec(pass)[0];
	i = pass.indexOf(phone);
	j = phone.length;
	if ((nondigit.test(pass[i-1])) && (nondigit.test(pass[i+j])))
	{
	    r += '<b>CAUTION: <font style="color:#840;">"'+phone+
	        '" looks like a telephone number!</font></b><br/>\n'; 
	    bioinfo++;
	}
    } 

    // Social Security number
    var pattern = /\d{3}\-\d{2}\-\d{4}/;
    if (pattern.test(pass))
    {
	ssn = pattern.exec(pass)[0];
	i = pass.indexOf(ssn);
	j = ssn.length;
	if ((nondigit.test(pass[i-1])) && (nondigit.test(pass[i+j])))
	{
	    r += '<b>CAUTION: <font style="color:#840;">"'+ssn+
		'" looks like a Social Security Number!</font></b><br/>\n';
	    bioinfo++;
	}
    } 

    // Date
    dyr = "((1[7-9]|20|\\')\\d{2}|\\d{2})";
    dyy = "(1[7-9]|20|\')\\d{2}";
    dmo = "(jan(uary|\\.?)?|feb(r(uary|\\.?)?|\\.?)?|mar(ch|\\.?)?|apr(il|\\.?)?|may|jun(e|\\.?)?|jul(y|\\.?)?|aug(ust|\\.?)?|sep(t(ember|\\.?)?|\\.?)?|oct(ober|\\.?)?|nov(ember|\\.?)?|dec(ember|\\.?)?)";
    dmm = "(0[1-9]|1[0-2])"
    ddd = "(0[1-9]|[1-2]\\d|3[0-1])"
    mmddyy = "(("+dmm+"\\ *"+ddd+"\\ *"+dyy+")|("+dmm+"\\ +"+ddd+"\\ +"+dyr+"))"
    ddmmyy = "(("+ddd+"\\ *"+dmm+"\\ *"+dyy+")|("+ddd+"\\ +"+dmm+"\\ +"+dyr+"))"
    yymmdd = "(("+dyy+"\\ *"+dmm+"\\ *"+ddd+")|("+dyr+"\\ +"+dmm+"\\ +"+ddd+"))"
    d0109 = "0?[1-9]";
    d1012 = "1[0-2]";
    d1331 = "(1[3-9]|2\\d|3[0-1])";
    d0131 = "(0[1-9]|[1-2]\\d?|3[0-1]?|[4-9])"
    d0112 = "(0[1-9]|1[0-2]?|[2-9])"
    dsep1 = "[\\/\\-\\.]";
    dsep2 = "\\,?\\ *";
    dsep3 = "[\\,\\ ]\\ *";
    dsep4 = "(\\/.*[\\-\\.]|\\-.*[\\/\\.]|\\..*[\\/\\-])"
    mdy = "("+d0109+dsep1+d0131+"|"+d1012+dsep1+d0131+"|"+d1331+dsep1+d0112+")"+dsep1+dyr
    md = d0112+dsep1+d0131;
    dmoy = "("+d0131+"\\ *)?"+dmo+dsep2+dyr;
    mody = dmo+"\\ *"+d0131+"("+dsep3+dyy+")?";
    moy = dmo+dsep2+dyy;
    ymd = dyy+dsep1+d0112+dsep1+d0131
    mpat = "("+moy+"|"+mody+"|"+dmoy+"|"+mmddyy+"|"+ddmmyy+"|"+yymmdd+"|"+ymd+"|"+mdy+"|"+md+")";
    var dpattern = new RegExp(mpat,"i");

    if (dpattern.test(plower))
    {
	date = dpattern.exec(plower)[0];
	j = date.length;
	k = 5;
	if (date.indexOf("/") > 0) k = 4;
	if (date.indexOf("-") > 0) k = 4;
	if (date.indexOf(".") > 0) k = 6;
	var spattern = new RegExp(dsep4);
	if (j >= k && !spattern.test(date))
	{
	    i = plower.indexOf(date);
	    oklo = 0;
	    okhi = 0;
	    if (i < 1) { oklo = 1; }
	    else if (plower.substr(i-1,2).match(/([^0-9][0-9]|[^a-z][a-z])/)) oklo = 1;
	    if (i +  j >= plower.length) { okhi = 1; }
	    else if (plower.substr(i+j-1,2).match(/([0-9][^0-9]|[a-z][^a-z])/)) okhi = 1;
	    if (oklo > 0 && okhi > 0)
	    {
		r += '<b>CAUTION: <font style="color:#840;">"'+pass.substr(i,j)+
		    '" looks like a date!</font></b><br/>\n';
		bioinfo++;
	    }
	}
    }

   // Calculate frequency chance
   if (pass.length > 1)
   {
      var c, aidx = 0, bits = 0, charSet;
      charSet = Math.log(Get_Charset_Size(pass)) / Math.log(2);
      aidx = Get_Index(plower.charAt(0));
      for (var b = 1; b < plower.length; b ++)
      {
	 var bidx = Get_Index(plower.charAt(b));
	 c = 1.0 - Frequency_Table[aidx * 27 + bidx];
	 bits += charSet * c * c;  // Squared = assmume they are good guessers
	 aidx = bidx;
      }

      // Arbitrary adjustments
      if (plower.length > 0)
      {
        if (bits > plower.length) bits *= (1-(commoncount*0.5/plower.length));	// Common words
        if (bits > plower.length) bits *= (1-(repeatcount*0.5/plower.length));	// Repeated sequences
        if (bits > plower.length) bits *= (1-(runcount*0.5/plower.length));	// Runs of one character
	if (bits > plower.length && uniques/charSet < 0.25)
	    bits *= (Math.sqrt(uniques*charSet)/charSet);	// Subset of character set actually used
      }

      if (bits < 28)
      {
         r += "<b>Strength: <font style='color:#c00;'>Very Weak</font></b> <br/> ";
	 r += "Try making your password longer, including CAPITALS, or <br/>";
	 r += "adding symbols.<br/>\n";
      }
      else if (bits < 36)
      {
         r += "<b>Strength: <font style='color:#c00;'>Weak</font></b> <br/>";
	 r += "Usually good enough for computer login passwords and to <br/>";
	 r += "keep out the average person.<br/>\n";
      }
      else if (bits < 60)
      {
         r += "<b>Strength: <font style='color:#808;'>Reasonable</font></b> <br/>";
	 r += "This password is fairly secure cryptographically and <br/>";
	 r += "skilled hackers may need some good computing power<br/>";
	 r += "to crack it.  (Depends greatly on implementation!)<br/>\n";
      }
      else if (bits < 128)
      {
         r += "<b>Strength: <font style='color:#060;'>Strong</font></b> <br/>";
	 r += "This password is typically good enough to safely guard <br/>";
	 r += "sensitive information like financial records.<br/>\n";
      }
      else
      {
         r += "<b>Strength: <font style='color:#048;'>Very Strong</font></b> <br/>";
	 r += "More often than not, this level of security is overkill.<br/>\n";
      }
      if (commoncount * 3 >= pass.length)
      {
         r += "<b>Note:</b> Common words account for "+Math.round(100*commoncount/pass.length)+
            "% of your password.<br/>";
         r += "This may make it easier to crack.<br/>\n";
      }
      if (repeatcount * 3 >= pass.length)
      {
         r += "<b>Note:</b> Repeated character sequences account for "+
            Math.min(100,Math.round(100*repeatcount/pass.length))+"% of<br/>";
         r += "your password. This may make it easier to crack.<br/>\n";
      }
      if (runcount * 3 >= pass.length)
      {
         r += "<b>Note:</b> Runs of a single character account for "+Math.round(100*runcount/pass.length)+
            "% of your<br/>";
         r += "password. This may make it easier to crack.<br/>\n";
      }
      if (bioinfo > 0)
      {
         r += "<b>Note:</b> You shouldn't include personal or related telephone<br/>";
         r += "numbers, social security numbers, dates or addresses in your<br/>";
         r += "passwords. They might be discovered by doing research about<br/>";
         r += "you, and this could make your password easier to crack.<br/>\n";
      }

      r += "<b>Length:</b> " + pass.length + ", \n";
      r += "<b>Entropy:</b> " + Math.round(bits) + " bits, \n";
      r += "<b>Charset Size:</b> " + Get_Charset_Size(pass) + " characters<br/>\n";
   }
   
   Set_Text(r);
   
   window.setTimeout('ShowStats();', 300);
}


function CheckIfLoaded()
{
   var s = "";
   if (! document.Common_Loaded)
   {
      s += "Loading common passwords...<br/>\n";
   }
   else if (! document.Common_Parsed)
   {
      if (! document.Common_Parsed_Started)
      {
         window.setTimeout('Parse_Common()', 50);
	 document.Common_Parsed_Started = 1;
      }
      s += "Parsing common passwords... " + 
         Common_List.length + "<br/>\n";
   }
   if (! document.Frequency_Loaded)
   {
      s += "Loading letter frequency table...<br/>\n";
   }
   else if (! document.Frequency_Parsed)
   {
      if (! document.Frequency_Parsed_Started)
      {
         window.setTimeout('Parse_Frequency()', 50);
	 document.Frequency_Parsed_Started = 1;
      }
      s += "Parsing frequency table... " + 
         Frequency_List.length + "<br/>\n";
   }
   if (s != "")
   {
      Set_Text(s + "Loading ...");
      window.setTimeout('CheckIfLoaded()', 200);
      return;
   }
   
   // Loaded. Do initialization thingies.
   Set_Text("Finished Loading.");
   window.setTimeout('ShowStats();', 1000);
}

window.setTimeout('CheckIfLoaded()', 100);
